from unittest.mock import patch
from mitosheet.mito_dash.v1.spreadsheet import Spreadsheet
from mitosheet.mito_dash.v1.mito_callback import mito_callback
from mitosheet.streamlit.v1 import RunnableAnalysis



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import re
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
import pandas as pd
from mitosheet.code_chunks.replace_code_chunk import ReplaceCodeChunk, convert_to_original_type_or_str

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.public.v3.types.bool import cast_string_to_bool
from mitosheet.errors import MitoError
from mitosheet.step_performers.column_steps.rename_column import rename_column_headers_in_state
from mitosheet.errors import make_invalid_replace_error
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID

class ReplaceStepPerformer(StepPerformer):
    """
    Allows you to replace a search value with a replace value in a sheet for both
    the values in the cells and the column headers.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'replace'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        search_value: str = get_param(params, 'search_value')
        replace_value: str = get_param(params, 'replace_value')
        column_ids: List[ColumnID] = get_param(params, 'column_ids')

        # We make a new state to modify it
        post_state = prev_state.copy() 

        pandas_start_time = perf_counter()
        df = post_state.dfs[sheet_index]

        # If no column_ids are specified, then we replace the values in all columns
        if column_ids is None or len(column_ids) == 0:
            column_ids = post_state.column_ids.get_column_ids(sheet_index)

        column_headers = post_state.column_ids.get_column_headers_by_ids(sheet_index, column_ids)

        # Selected columns is the dataframe with only the columns we want to replace values in
        df_only_selected_columns = df[column_headers]

        # Raise an error if the pandas version is too old to use timedelta with replace.
        if (any(df_only_selected_columns.dtypes == 'timedelta') and pd.__version__ < 1.4):
            raise MitoError(
                'version_error',
                'Pandas version error',
                'This version of pandas doesn\'t support replacing values in timedelta columns. Please upgrade to pandas 1.2 or later.',
            )
        
        try:
            search_value_regex = f'(?i){search_value}'
            # Special case for boolean columns because when we convert to string, the values are 
            # all converted back to bool as True - even if the value is "False". 
            if any(df_only_selected_columns.dtypes == 'bool'):
                bool_columns = df_only_selected_columns.select_dtypes(include='bool')
                non_bool_columns = df_only_selected_columns.select_dtypes(exclude='bool')
                df_only_selected_columns[bool_columns.columns] = bool_columns.astype(str).replace(search_value_regex, replace_value, regex=True).applymap(cast_string_to_bool).astype(bool)
                df_only_selected_columns[non_bool_columns.columns] = non_bool_columns.astype(str).replace(search_value_regex, replace_value, regex=True).astype(non_bool_columns.dtypes.to_dict())
            else:
                df_only_selected_columns = df_only_selected_columns.astype(str).replace(search_value_regex, replace_value, regex=True).astype(df_only_selected_columns.dtypes.to_dict())

            # Then, we replace the column headers in the state column_ids object
            # We convert the column headers to strings because the column headers can be any type
            new_columns = [convert_to_original_type_or_str(re.sub(search_value_regex, replace_value, str(column)), type(column)) for column in df_only_selected_columns.columns]

            # Update the column headers in the state column_ids object
            for old_column_name, new_column_name in zip(df_only_selected_columns.columns, new_columns):
                # If the column name didn't change, then we don't need to do anything
                if old_column_name == new_column_name:
                    continue
                column_id = post_state.column_ids.get_column_id_by_header(sheet_index, old_column_name)
                post_state.column_ids.set_column_header(sheet_index, column_id, new_column_name)

            # Replace the selected columns in the actual dataframe
            df[column_headers] = df_only_selected_columns

            # Finally, we replace the column headers in the dataframe
            df.rename(columns=dict(zip(column_headers, new_columns)), inplace=True)
            post_state.dfs[sheet_index] = df
        except Exception as e:
            raise make_invalid_replace_error(search_value, replace_value)
        pandas_processing_time = perf_counter() - pandas_start_time
        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ReplaceCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_ids'),
                get_param(params, 'search_value'),
                get_param(params, 'replace_value'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}
    


import gc
from io import StringIO
import json
import time
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Union, Tuple

from mitosheet.mito_dash.v1.spreadsheet import Spreadsheet, WRONG_CALLBACK_ERROR_MESSAGE

WRONG_CALLBACK_ERROR_MESSAGE_FIRST_LINE = WRONG_CALLBACK_ERROR_MESSAGE.split('\n')[0]

try:

    from dash import callback
        
    def get_component_with_id(id: str) -> Optional[Spreadsheet]:
        components = [
            obj for obj in gc.get_objects()
            if isinstance(obj, Spreadsheet) and getattr(obj, 'mito_id', None) == id
        ]

        if len(components) > 0:
            return components[0]
        else:
            return None        

    def mito_callback(*args, **kwargs):
        """
        This is a replacement for the @callback decorator that improves the developer experience
        with using the Mito Spreadsheet component. 

        When using the @mito_callback decorator to subscribe to changes in the `spreadsheet_result` or
        `spreadsheet_selection` properties of a Mito Spreadsheet, you get the actual SpreadsheetResult
        object - not some JSON string you need to parse and work with.

        To see the documentation for this decorator, see the @mito_callback decorator documentation:
        https://docs.trymito.io/mito-for-dash/api-reference#callback-props-and-types
        """
        def function_wrapper(original_function):
            # TODO: do we need an @wraps
            def new_function(*_args, **_kwargs):
                new_args = list(_args)

                # TODO: do we need to handle kwargs? Probably. But we will do that in the future...

                for index, arg in enumerate(new_args):

                    # We go through all the arguments passed to the function, and:
                    # 1) find those that are Mito spreadsheet_result/spreadsheet_selection error strings
                    # 2) Replace them with the the actual spreadsheet results
                    # As this is a dramatically better Spreadsheet experience

                    if isinstance(arg, str) and arg.startswith(WRONG_CALLBACK_ERROR_MESSAGE_FIRST_LINE):
                        # Get the ID from the final line 
                        id = arg.split('\n')[-1].strip()

                        spreadsheet = get_component_with_id(id)
                        if spreadsheet is None:
                            # TODO: use a more dash exception?
                            raise Exception(f"Could not find spreadsheet with id {id}")
                        
                        if 'spreadsheet_result' in arg:
                            # If the user is getting the result, give them the result
                            new_args[index] = spreadsheet.get_result()
                        else:
                            # If they are just getting the selection, just get the current selection
                            new_args[index] = spreadsheet.get_result().selection()


                return original_function(*new_args, **_kwargs)
            
            return callback(*args, **kwargs)(new_function)
        
        return function_wrapper
        
except ImportError:

    def mito_callback(*args, **kwargs): # type: ignore
        raise Exception("You must install dash to use the @mito_callback decorator component")



import streamlit as st
import pandas as pd 
from mitosheet.streamlit.v1 import spreadsheet
from mitosheet.extensions.v1 import ColumnHeader

st.set_page_config(layout="wide")

def ADD_ONE():
    return 1

def get_data_from_database(first_param: int, second_param: float) -> pd.DataFrame:
    return df

def do_import_with_a_really_long_name(x: int, y: float) -> pd.DataFrame:
    return df + x + y

def add_number_to_df(df: pd.DataFrame, column_header: ColumnHeader, addition: int) -> pd.DataFrame:
    df[column_header] = df[column_header] + addition
    return df

def add_diff_number_to_df_and_this_is_a_long_name(df: pd.DataFrame, number_to_add_and_long_param_name: int, boolean_with_long_name: bool, string_with_long_name: str) -> pd.DataFrame:
    return df + number_to_add_and_long_param_name


df = pd.DataFrame({ 'A': [1, 2, 3], 'B': [4, 5, 6] })
analysis = spreadsheet(
    df,
    df_names=['df'],
    import_folder='datasets',
    editors=[add_number_to_df],
    importers=[get_data_from_database, do_import_with_a_really_long_name],
    return_type='analysis'
)

st.write(analysis.get_param_metadata())
updated_metadata = {}
for param in analysis.get_param_metadata():
    new_param = None
    if param['subtype'] in ['file_name_export_excel', 'file_name_export_csv']:
        new_param = st.text_input(param['name'], value=param['original_value'])
    elif param['subtype'] in ['file_name_import_excel', 'file_name_import_csv']:
        new_param = st.file_uploader(param['name'])
    if new_param is not None:
        updated_metadata[param['name']] = new_param


run = st.button('Run')
if run:
    result = analysis.run(df, **updated_metadata)
    st.write(result)




import gc
from io import StringIO
import json
import time
from queue import Queue
from typing import Any, Callable, Dict, List, Optional, Union, Tuple
from unittest.mock import patch

import pandas as pd
from mitosheet.mito_backend import MitoBackend
from mitosheet.selectionUtils import get_selected_element
from mitosheet.utils import get_new_id, get_new_id
from mitosheet.types import CodeOptions, MitoTheme, MitoFrontendIndexAndSelections, ParamMetadata
from mitosheet.streamlit.v1 import RunnableAnalysis


class SpreadsheetResult():

    def __init__(
        self, 
        dfs: List[pd.DataFrame],
        code: List[str],
        fully_parameterized_function: str,
        param_metadata: List[ParamMetadata],
        code_options: CodeOptions,
        index_and_selections: Optional[MitoFrontendIndexAndSelections]=None
    ):
        self.__dfs = dfs
        self.__code = code
        self.__index_and_selections = index_and_selections
        self.__fully_parameterized_function = fully_parameterized_function
        self.__param_metadata = param_metadata
        self.__code_options = code_options

    def dfs(self) -> List[pd.DataFrame]:
        return self.__dfs
    
    def code(self) -> str:
        return "\n".join(self.__code)
    
    def selection(self) -> Optional[Union[pd.DataFrame, pd.Series]]:
        return get_selected_element(self.__dfs, self.__index_and_selections)
    
    def analysis(self) -> RunnableAnalysis:
        return RunnableAnalysis(self.code(), self.__code_options, self.__fully_parameterized_function, self.__param_metadata)
    
WRONG_CALLBACK_ERROR_MESSAGE = """Error: Registering a callback with an Input or State referencing a Mito Spreadsheet requires using the @mito_callback decorator, rather than the @callback decorator.

To get the proper value from the {prop_name}, change your callback to use the @mito_callback decorator instead of the @callback decorator.

See more: https://docs.trymito.io/mito-for-dash/api-reference#callback-props-and-types

{num_messages}
{prop_name}
{id}"""
         
try:
    from dash.development.base_component import Component
    from dash import Input, Output, callback, State


    class Spreadsheet(Component):

        _children_props: List[str] = []
        _base_nodes = ['children']
        _namespace = 'dash_spreadsheet_v1'
        _type = 'MitoDashWrapper'
        _prop_names = ['id', 'all_json', 'data', 'import_folder', 'spreadsheet_result', 'spreadsheet_selection']
        _valid_wildcard_attributes: List[str] = []
        available_properties = ['id', 'all_json', 'data', 'import_folder']
        available_wildcard_properties: List[str] = []

        def __init__(
                self, 
                *args: Union[pd.DataFrame, str, None],
                id: str,
                import_folder: Optional[str]=None,
                code_options: Optional[CodeOptions]=None,
                df_names: Optional[List[str]]=None,
                sheet_functions: Optional[List[Callable]]=None, 
                importers: Optional[List[Callable]]=None,
                editors: Optional[List[Callable]]=None,
                theme: Optional[MitoTheme]=None,
                track_selection: bool=False,
                
        ):     
            self.mito_id = id
            # Note: num_messages must be ever increasing, so that whenever we get a new message, the spreadsheet_result
            # and spreadsheet_selection error strings change. This way, we can correctly trigger callbacks that correspond
            # to these values in all cases
            self.num_messages=0
            self._set_new_mito_backend(
                *args, 
                import_folder=import_folder, 
                code_options=code_options,
                df_names=df_names,
                sheet_functions=sheet_functions,
                importers=importers,
                editors=editors,
                theme=theme
            )

            self.track_selection = track_selection

            super(Spreadsheet, self).__init__(
                id=id,
                all_json=self.get_all_json(),
            )

            # We save the unprocessed messages in a list -- so that we can process them
            # in the callback in the order that they were received -- without them interrupting
            # eachother and having to deal with race conditions
            self.unprocessed_messages: Any = Queue()
            self.processing_messages = False

            self.index_and_selections: Optional[MitoFrontendIndexAndSelections] = None

            # Make sure to save import-folder and code-options as attributes, so if we need
            # to recreate the backend, we can do so
            self.import_folder = import_folder
            self.code_options = code_options
            self.df_names = df_names
            self.sheet_functions = sheet_functions
            self.importers = importers
            self.editors = editors
            self.theme = theme

            @callback(
                Output(self.mito_id, 'all_json', allow_duplicate=True), 
                Output(self.mito_id, 'spreadsheet_result', allow_duplicate=True), 
                Input(self.mito_id, 'message'), prevent_initial_call=True
            )
            def handle_message(msg):
                self.num_messages += 1

                self.unprocessed_messages.put(msg)
                self.process_single_message()
                
                self.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=self.num_messages, id=self.mito_id)
                return self.get_all_json(), self.spreadsheet_result
            
            # Because this has a performance impact, we only register this callback if
            # the user actually uses the track_selection parameter
            # TODO: improve the selection error message in this case...
            if track_selection:
                @callback(
                    Output(self.mito_id, 'all_json', allow_duplicate=True), 
                    Output(self.mito_id, 'spreadsheet_selection', allow_duplicate=True), 
                    Input(self.mito_id, 'index_and_selections'), prevent_initial_call=True
                )
                def handle_selection_change(index_and_selections):
                    self.num_messages += 1

                    self.index_and_selections = index_and_selections
                    
                    self.spreadsheet_selection = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_selection', num_messages=self.num_messages, id=self.mito_id)
                    return self.get_all_json(), self.spreadsheet_selection

            @callback(
                Output(self.mito_id, 'all_json', allow_duplicate=True), 
                Output(self.mito_id, 'spreadsheet_result', allow_duplicate=True), 
                Output(self.mito_id, 'spreadsheet_selection', allow_duplicate=True), 
                Input(self.mito_id, 'data'), 
                prevent_initial_call=True
            )
            def handle_data_change_data(data):
                
                self._set_new_mito_backend(
                    data, 
                    import_folder=self.import_folder, 
                    code_options=self.code_options,
                    df_names=self.df_names,
                    sheet_functions=self.sheet_functions,
                    importers=self.importers,
                    editors=self.editors,
                    theme=self.theme
                )

                
                return self.get_all_json(), self.spreadsheet_result, self.spreadsheet_selection
            
        def _set_new_mito_backend(
                self, 
                *args: Union[pd.DataFrame, str, None], 
                import_folder: Optional[str]=None,
                code_options: Optional[CodeOptions]=None,
                df_names: Optional[List[str]]=None,
                sheet_functions: Optional[List[Callable]]=None, 
                importers: Optional[List[Callable]]=None,
                editors: Optional[List[Callable]]=None,
                theme: Optional[MitoTheme]=None
            ) -> None:
            """
            Called when the component is created, or when the input data is changed.
            """
            self.mito_frontend_key = get_new_id()
            self.mito_backend = MitoBackend(
                *args, 
                import_folder=import_folder, 
                code_options=code_options,
                user_defined_functions=sheet_functions,
                user_defined_importers=importers,
                user_defined_editors=editors,
                theme=theme
            )
            self.responses: List[Dict[str, Any]] = []
            def send(response):
                self.responses.append(response)
            self.mito_backend.mito_send = send

            # If there are any df_names, then we send them to the backend as well. 
            # TODO: we should be able to pass this directly to the backend
            if df_names is not None and len(df_names) > 0:
                self.mito_backend.receive_message(
                    {
                        'event': 'update_event',
                        'id': get_new_id(),
                        'type': 'args_update',
                        'params': {
                            'args': df_names
                        },
                    }
                )

            # If you use the @callback decorator with spreadsheet_result or spreadsheet_selection, users will get these helpful error messages that 
            # will tell them what is wrong with their approach - namely, that they need to use `@mito_callback` instead. Note that this includes:
            # 1.    The num_messages so it triggers callbacks correct in the all cases
            # 2.    The prop name and the id of the spreadsheet so the `@mito_callback` function can inspect this string and replace this arg with the
            #       actual value the user really wants
            self.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=self.num_messages, id=self.mito_id)
            self.spreadsheet_selection = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_selection', num_messages=self.num_messages, id=self.mito_id)

                
        def process_single_message(self):

            # If we are already processing messages -- then wait until it is
            if self.processing_messages:
                while self.processing_messages:
                    time.sleep(0.1)
                
            # Otherwise, set the processing flag to true
            self.processing_messages = True

            # Process all the messages in the queue
            try:
                if not self.unprocessed_messages.empty():
                    value = self.unprocessed_messages.get()
                    self.mito_backend.receive_message(value)
            except:
                # Make sure we always set the processing flag to false
                pass
                
            # Set the processing flag to false
            self.processing_messages = False

            self.spreadsheet_result = WRONG_CALLBACK_ERROR_MESSAGE.format(prop_name='spreadsheet_result', num_messages=self.num_messages, id=self.mito_id)
            
            
        def get_all_json(self) -> str:
            return json.dumps({
                **self.mito_backend.get_shared_state_variables(),
                'responses_json': json.dumps(self.responses),
                'key': self.mito_frontend_key,
                'track_selection': self.track_selection
            })
        
        def get_result(self):
            return SpreadsheetResult(
                dfs=self.mito_backend.steps_manager.dfs,
                code=self.mito_backend.steps_manager.code(),
                index_and_selections=self.index_and_selections,
                fully_parameterized_function=self.mito_backend.fully_parameterized_function,
                param_metadata=self.mito_backend.param_metadata,
                code_options=self.code_options
            )
        
except ImportError:

    class Spreadsheet(): # type: ignore
        def __init__(self, *args, **kwargs):
            raise Exception("You must install dash to use the Spreadsheet component")